@@ -1,4 +1,4 @@
-// Copyright 2019 Graham Clark. All rights reserved.  Use of this source
+// Copyright 2019-2020 Graham Clark. All rights reserved.  Use of this source
 // code is governed by the MIT license that can be found in the LICENSE
 // file.
 
@@ -9,12 +9,10 @@ import (
 	"io"
 	"os"
 	"os/exec"
-	"runtime"
 	"strings"
 	"sync"
 
 	"github.com/gcla/termshark/v2"
-	"github.com/pkg/errors"
 	log "github.com/sirupsen/logrus"
 )
 
@@ -47,6 +45,7 @@ func (c *Command) Start() error {
 	c.Lock()
 	defer c.Unlock()
 	c.Cmd.Stderr = log.StandardLogger().Writer()
+	c.PutInNewGroupOnUnix()
 	res := c.Cmd.Start()
 	return res
 }
@@ -78,19 +77,6 @@ func (c *Command) Close() error {
 	return nil
 }
 
-func (c *Command) Kill() error {
-	c.Lock()
-	defer c.Unlock()
-	if c.Cmd.Process == nil {
-		return errors.WithStack(ProcessNotStarted{Command: c.Cmd})
-	}
-	if runtime.GOOS == "windows" {
-		return c.Cmd.Process.Kill()
-	} else {
-		return c.Cmd.Process.Signal(os.Interrupt)
-	}
-}
-
 func (c *Command) Pid() int {
 	c.Lock()
 	defer c.Unlock()
@@ -122,12 +108,29 @@ func MakeCommands(decodeAs []string, args []string, pdml []string, psml []string
 
 var _ ILoaderCmds = Commands{}
 
-func (c Commands) Iface(iface string, captureFilter string, tmpfile string) IBasicCommand {
-	args := []string{"-P", "-i", iface, "-w", tmpfile}
+func (c Commands) Iface(ifaces []string, captureFilter string, tmpfile string) IBasicCommand {
+	args := make([]string, 0)
+	for _, iface := range ifaces {
+		args = append(args, "-i", iface)
+	}
+	args = append(args, "-w", tmpfile)
 	if captureFilter != "" {
 		args = append(args, "-f", captureFilter)
 	}
-	return &Command{Cmd: exec.Command(termshark.DumpcapBin(), args...)}
+	res := &Command{
+		Cmd: exec.Command(termshark.CaptureBin(), args...),
+	}
+	// This tells termshark to start in a special capture mode. It allows termshark
+	// to run itself like this:
+	//
+	// termshark -i eth0 -w foo.pcap
+	//
+	// which will then run dumpcap and if that fails, tshark. The idea
+	// is to use the most specialized/efficient capture method if that
+	// works, but fall back to tshark if needed e.g. for randpkt, sshcapture, etc
+	// (extcap interfaces).
+	res.Cmd.Env = append(os.Environ(), "TERMSHARK_CAPTURE_MODE=1")
+	return res
 }
 
 func (c Commands) Tail(tmpfile string) ITailCommand {
@@ -185,7 +188,7 @@ func (c Commands) Psml(pcap interface{}, displayFilter string) IPcapCommand {
 
 func (c Commands) Pcap(pcap string, displayFilter string) IPcapCommand {
 	// need to use stdout and -w - otherwise, tshark writes one-line text output
-	args := []string{"-F", "pcap", "-r", pcap, "-w", "-"}
+	args := []string{"-r", pcap, "-x"}
 	if displayFilter != "" {
 		args = append(args, "-Y", displayFilter)
 	}
