@@ -1,4 +1,4 @@
-// Copyright 2019 Graham Clark. All rights reserved.  Use of this source
+// Copyright 2019-2020 Graham Clark. All rights reserved.  Use of this source
 // code is governed by the MIT license that can be found in the LICENSE
 // file.
 
@@ -45,7 +45,7 @@ import (
 	"github.com/gcla/termshark/v2"
 	"github.com/gcla/termshark/v2/pcap"
 	"github.com/gcla/termshark/v2/pdmltree"
-	"github.com/gcla/termshark/v2/psmltable"
+	"github.com/gcla/termshark/v2/psmlmodel"
 	"github.com/gcla/termshark/v2/system"
 	"github.com/gcla/termshark/v2/ui/menuutil"
 	"github.com/gcla/termshark/v2/widgets"
@@ -57,6 +57,7 @@ import (
 	"github.com/gcla/termshark/v2/widgets/hexdumper2"
 	"github.com/gcla/termshark/v2/widgets/ifwidget"
 	"github.com/gcla/termshark/v2/widgets/resizable"
+	"github.com/gcla/termshark/v2/widgets/rossshark"
 	"github.com/gcla/termshark/v2/widgets/withscrollbar"
 	"github.com/gdamore/tcell"
 	lru "github.com/hashicorp/golang-lru"
@@ -71,7 +72,7 @@ var Goroutinewg *sync.WaitGroup
 // Global so that we can change the displayed packet in the struct view, etc
 // test
 var appViewNoKeys *holder.Widget
-var appView *appkeys.KeyWidget
+var appView *holder.Widget
 var mainViewNoKeys *holder.Widget
 var mainView *appkeys.KeyWidget
 var pleaseWaitSpinner *spinner.Widget
@@ -106,6 +107,7 @@ var generalMenu *menu.Widget
 var analysisMenu *menu.Widget
 var savedMenu *menu.Widget
 var FilterWidget *filter.Widget
+var Fin *rossshark.Widget
 var CopyModeWidget gowid.IWidget
 var CopyModePredicate ifwidget.Predicate
 var openMenuSite *menu.SiteWidget
@@ -134,7 +136,7 @@ var colSpace *gowid.ContainerWidget
 
 var curPacketStructWidget *copymodetree.Widget
 var packetHexWidgets *lru.Cache
-var packetListView *rowFocusTableWidget
+var packetListView *psmlTableRowWidget
 
 var Loadingw gowid.IWidget    // "loading..."
 var MissingMsgw gowid.IWidget // centered, holding singlePacketViewMsgHolder
@@ -502,22 +504,46 @@ func (w *selectedComposite) SubWidget() gowid.IWidget {
 
 //======================================================================
 
+// An ugly interface that captures what sort of type will be suitable
+// as a table widget to which a row focus can be applied.
+type iRowFocusTableWidgetNeeds interface {
+	gowid.IWidget
+	list.IBoundedWalker
+	table.IFocus
+	table.IGoToMiddle
+	table.ISetFocus
+	list.IWalkerHome
+	list.IWalkerEnd
+	SetModel(table.IModel, gowid.IApp)
+	FocusXY() (table.Coords, error)
+	SetFocusXY(gowid.IApp, table.Coords)
+	Lower() *table.ListWithPreferedColumn
+	SetFocusOnData(app gowid.IApp) bool
+	OnFocusChanged(f gowid.IWidgetChangedCallback)
+}
+
 // rowFocusTableWidget provides a table that highlights the selected row or
 // focused row.
 type rowFocusTableWidget struct {
-	// set to true after the first time we move focus from the table header to the data. We do this
-	// once and that this happens quickly, but then assume the user might want to move back to the
-	// table header manually, and it would be strange if the table keeps jumping back to the data...
-	didFirstAutoFocus bool
-	*table.BoundedWidget
-	colors []pcap.PacketColors
+	iRowFocusTableWidgetNeeds
+	rowSelected string
+	rowFocus    string
+}
+
+func NewRowFocusTableWidget(w iRowFocusTableWidgetNeeds, rs string, rf string) *rowFocusTableWidget {
+	res := &rowFocusTableWidget{
+		iRowFocusTableWidgetNeeds: w,
+		rowSelected:               rs,
+		rowFocus:                  rf,
+	}
+	res.Lower().IWidget = list.NewBounded(res)
+	return res
 }
 
 var _ gowid.IWidget = (*rowFocusTableWidget)(nil)
-var _ gowid.IComposite = (*rowFocusTableWidget)(nil)
 
 func (t *rowFocusTableWidget) SubWidget() gowid.IWidget {
-	return t.BoundedWidget
+	return t.iRowFocusTableWidgetNeeds
 }
 
 func (t *rowFocusTableWidget) InvertedModel() table.IInvertible {
@@ -525,7 +551,7 @@ func (t *rowFocusTableWidget) InvertedModel() table.IInvertible {
 }
 
 func (t *rowFocusTableWidget) Rows() int {
-	return t.Widget.Model().(table.IBoundedModel).Rows()
+	return t.Model().(table.IBoundedModel).Rows()
 }
 
 // Implement withscrollbar.IScrollValues
@@ -540,32 +566,32 @@ func (t *rowFocusTableWidget) ScrollPosition() int {
 
 func (t *rowFocusTableWidget) Up(lines int, size gowid.IRenderSize, app gowid.IApp) {
 	for i := 0; i < lines; i++ {
-		t.Widget.UserInput(tcell.NewEventKey(tcell.KeyUp, ' ', tcell.ModNone), size, gowid.Focused, app)
+		t.UserInput(tcell.NewEventKey(tcell.KeyUp, ' ', tcell.ModNone), size, gowid.Focused, app)
 	}
 }
 
 func (t *rowFocusTableWidget) Down(lines int, size gowid.IRenderSize, app gowid.IApp) {
 	for i := 0; i < lines; i++ {
-		t.Widget.UserInput(tcell.NewEventKey(tcell.KeyDown, ' ', tcell.ModNone), size, gowid.Focused, app)
+		t.UserInput(tcell.NewEventKey(tcell.KeyDown, ' ', tcell.ModNone), size, gowid.Focused, app)
 	}
 }
 
 func (t *rowFocusTableWidget) UpPage(num int, size gowid.IRenderSize, app gowid.IApp) {
 	for i := 0; i < num; i++ {
-		t.Widget.UserInput(tcell.NewEventKey(tcell.KeyPgUp, ' ', tcell.ModNone), size, gowid.Focused, app)
+		t.UserInput(tcell.NewEventKey(tcell.KeyPgUp, ' ', tcell.ModNone), size, gowid.Focused, app)
 	}
 }
 
 func (t *rowFocusTableWidget) DownPage(num int, size gowid.IRenderSize, app gowid.IApp) {
 	for i := 0; i < num; i++ {
-		t.Widget.UserInput(tcell.NewEventKey(tcell.KeyPgDn, ' ', tcell.ModNone), size, gowid.Focused, app)
+		t.UserInput(tcell.NewEventKey(tcell.KeyPgDn, ' ', tcell.ModNone), size, gowid.Focused, app)
 	}
 }
 
 // list.IWalker
 func (t *rowFocusTableWidget) At(lpos list.IWalkerPosition) gowid.IWidget {
 	pos := int(lpos.(table.Position))
-	w := t.Widget.AtRow(pos)
+	w := t.AtRow(pos)
 	if w == nil {
 		return nil
 	}
@@ -573,17 +599,11 @@ func (t *rowFocusTableWidget) At(lpos list.IWalkerPosition) gowid.IWidget {
 	// Composite so it passes through prefered column
 	var res gowid.IWidget = &selectedComposite{
 		Widget: isselected.New(w,
-			styled.New(w, gowid.MakePaletteRef("pkt-list-row-selected")),
-			styled.New(w, gowid.MakePaletteRef("pkt-list-row-focus")),
+			styled.New(w, gowid.MakePaletteRef(t.rowSelected)),
+			styled.New(w, gowid.MakePaletteRef(t.rowFocus)),
 		),
 	}
 
-	if pos >= 0 && PacketColors {
-		res = styled.New(res,
-			gowid.MakePaletteEntry(t.colors[pos].FG, t.colors[pos].BG),
-		)
-	}
-
 	return res
 }
 
@@ -595,6 +615,49 @@ func (t *rowFocusTableWidget) Focus() list.IWalkerPosition {
 
 //======================================================================
 
+// A rowFocusTableWidget that adds colors to rows
+type psmlTableRowWidget struct {
+	*rowFocusTableWidget
+	// set to true after the first time we move focus from the table header to the data. We do this
+	// once and that this happens quickly, but then assume the user might want to move back to the
+	// table header manually, and it would be strange if the table keeps jumping back to the data...
+	didFirstAutoFocus bool
+	colors            []pcap.PacketColors
+}
+
+func NewPsmlTableRowWidget(w *rowFocusTableWidget, c []pcap.PacketColors) *psmlTableRowWidget {
+	res := &psmlTableRowWidget{
+		rowFocusTableWidget: w,
+		colors:              c,
+	}
+	res.Lower().IWidget = list.NewBounded(res)
+	return res
+}
+
+func (t *psmlTableRowWidget) At(lpos list.IWalkerPosition) gowid.IWidget {
+	res := t.rowFocusTableWidget.At(lpos)
+	if res == nil {
+		return nil
+	}
+	pos := int(lpos.(table.Position))
+
+	// Check the color array length because it might not yet be adequately
+	// populated from the arriving psml.
+	if pos >= 0 && PacketColors && pos < len(t.colors) {
+		res = styled.New(res,
+			gowid.MakePaletteEntry(t.colors[pos].FG, t.colors[pos].BG),
+		)
+	}
+
+	return res
+}
+
+func (t *psmlTableRowWidget) Focus() list.IWalkerPosition {
+	return table.Focus(t)
+}
+
+//======================================================================
+
 type pleaseWaitCallbacks struct {
 	w    *spinner.Widget
 	app  gowid.IApp
@@ -752,13 +815,8 @@ func openResultsAfterCopy(tmplName string, tocopy string, app gowid.IApp) {
 	termshark.CopyCommand(strings.NewReader(tocopy), v)
 }
 
-func openCopyChoices(copyLen int, app gowid.IApp) {
+func processCopyChoices(copyLen int, app gowid.IApp) {
 	var cc *dialog.Widget
-	maximizer := &dialog.Maximizer{}
-
-	clips := app.Clips()
-
-	cws := make([]gowid.IWidget, 0, len(clips))
 
 	copyCmd := termshark.ConfStringSlice(
 		"main.copy-command",
@@ -770,6 +828,24 @@ func openCopyChoices(copyLen int, app gowid.IApp) {
 		return
 	}
 
+	clips := app.Clips()
+
+	// No need to display a choice dialog with one choice - just copy right away
+	if len(clips) == 1 {
+		app.InCopyMode(false)
+		termshark.CopyCommand(strings.NewReader(clips[0].ClipValue()), userCopiedCallbacks{
+			app:     app,
+			copyCmd: copyCmd,
+			pleaseWaitCallbacks: &pleaseWaitCallbacks{
+				w:   pleaseWaitSpinner,
+				app: app,
+			},
+		})
+		return
+	}
+
+	cws := make([]gowid.IWidget, 0, len(clips))
+
 	for _, clip := range clips {
 		c2 := clip
 		lbl := text.New(clip.ClipName() + ":")
@@ -818,21 +894,6 @@ func openCopyChoices(copyLen int, app gowid.IApp) {
 
 	var view1 gowid.IWidget = pile.NewFlow(text.New("Select option to copy:"), divider.NewUnicode(), view2)
 
-	view1 = appkeys.New(
-		view1,
-		func(ev *tcell.EventKey, app gowid.IApp) bool {
-			if ev.Rune() == 'z' { // maximize/unmaximize
-				if maximizer.Maxed {
-					maximizer.Unmaximize(cc, app)
-				} else {
-					maximizer.Maximize(cc, app)
-				}
-				return true
-			}
-			return false
-		},
-	)
-
 	cc = dialog.New(view1,
 		dialog.Options{
 			Buttons:         dialog.CloseOnly,
@@ -840,6 +901,7 @@ func openCopyChoices(copyLen int, app gowid.IApp) {
 			BackgroundStyle: gowid.MakePaletteRef("dialog"),
 			BorderStyle:     gowid.MakePaletteRef("dialog"),
 			ButtonStyle:     gowid.MakePaletteRef("dialog-buttons"),
+			FocusOnWidget:   true,
 		},
 	)
 
@@ -1051,6 +1113,8 @@ func reallyClear(app gowid.IApp) {
 							pcap.HandlerList{
 								MakePacketViewUpdater(app),
 								MakeUpdateCurrentCaptureInTitle(app),
+								ManageStreamCache{},
+								ManageCapinfoCache{},
 							},
 						)
 					},
@@ -1165,7 +1229,7 @@ func copyModeExitKeysClipped(evk *tcell.EventKey, copyLen int, app gowid.IApp) b
 		case tcell.KeyEscape:
 			app.InCopyMode(false)
 		case tcell.KeyCtrlC:
-			openCopyChoices(copyLen, app)
+			processCopyChoices(copyLen, app)
 		case tcell.KeyRight:
 			cl := app.CopyModeClaimedAt()
 			app.CopyModeClaimedAt(cl + 1)
@@ -1425,7 +1489,14 @@ func setLowerWidgets(app gowid.IApp) {
 
 			hex := getHexWidgetToDisplay(row)
 			if hex != nil {
-				sw1 = enableselected.New(hex)
+				sw1 = enableselected.New(
+					withscrollbar.New(
+						hex,
+						withscrollbar.Options{
+							HideIfContentFits: true,
+						},
+					),
+				)
 			}
 			str := getStructWidgetToDisplay(row, app)
 			if str != nil {
@@ -1452,7 +1523,7 @@ func setLowerWidgets(app gowid.IApp) {
 
 }
 
-func makePacketListModel(psml psmlInfo, app gowid.IApp) *psmltable.Model {
+func makePacketListModel(psml psmlInfo, app gowid.IApp) *psmlmodel.Model {
 	packetPsmlTableModel := table.NewSimpleModel(
 		psml.PsmlHeaders(),
 		psml.PsmlData(),
@@ -1479,7 +1550,7 @@ func makePacketListModel(psml psmlInfo, app gowid.IApp) *psmltable.Model {
 		},
 	)
 
-	expandingModel := psmltable.New(
+	expandingModel := psmlmodel.New(
 		packetPsmlTableModel,
 		gowid.MakePaletteRef("pkt-list-row-focus"),
 	)
@@ -1526,12 +1597,15 @@ func setPacketListWidgets(psml psmlInfo, app gowid.IApp) {
 	expandingModel := makePacketListModel(psml, app)
 
 	packetListTable = &table.BoundedWidget{Widget: table.New(expandingModel)}
-	packetListView = &rowFocusTableWidget{
-		BoundedWidget: packetListTable,
-		colors:        psml.PsmlColors(),
-	}
+	packetListView = NewPsmlTableRowWidget(
+		NewRowFocusTableWidget(
+			packetListTable,
+			"pkt-list-row-selected",
+			"pkt-list-row-focus",
+		),
+		psml.PsmlColors(),
+	)
 
-	packetListView.Lower().IWidget = list.NewBounded(packetListView)
 	packetListView.OnFocusChanged(gowid.MakeWidgetCallback("cb", func(app gowid.IApp, w gowid.IWidget) {
 		fxy, err := packetListView.FocusXY()
 		if err != nil {
@@ -1909,6 +1983,7 @@ func RequestLoadPcapWithCheck(pcapf string, displayFilter string, app gowid.IApp
 				MakePacketViewUpdater(app),
 				MakeUpdateCurrentCaptureInTitle(app),
 				ManageStreamCache{},
+				ManageCapinfoCache{},
 			},
 		)
 	}
@@ -2373,6 +2448,14 @@ func Build() (*gowid.App, error) {
 	}))
 
 	analysisMenuItems := []menuutil.SimpleMenuItem{
+		menuutil.SimpleMenuItem{
+			Txt: "Capture file properties",
+			Key: gowid.MakeKey('p'),
+			CB: func(app gowid.IApp, w gowid.IWidget) {
+				analysisMenu.Close(app)
+				startCapinfo(app)
+			},
+		},
 		menuutil.SimpleMenuItem{
 			Txt: "Reassemble stream",
 			Key: gowid.MakeKey('f'),
@@ -2381,6 +2464,14 @@ func Build() (*gowid.App, error) {
 				startStreamReassembly(app)
 			},
 		},
+		menuutil.SimpleMenuItem{
+			Txt: "Conversations",
+			Key: gowid.MakeKey('c'),
+			CB: func(app gowid.IApp, w gowid.IWidget) {
+				analysisMenu.Close(app)
+				openConvsUi(app)
+			},
+		},
 	}
 
 	analysisMenuListBox := menuutil.MakeMenuWithHotKeys(analysisMenuItems)
@@ -2504,6 +2595,7 @@ func Build() (*gowid.App, error) {
 				MakeSaveRecents("", FilterWidget.Value(), app),
 				MakePacketViewUpdater(app),
 				ManageStreamCache{},
+				ManageCapinfoCache{},
 			},
 		)
 	})
@@ -2867,6 +2959,7 @@ func Build() (*gowid.App, error) {
 	menuPathMax = []interface{}{0, 6}
 
 	buildStreamUi()
+	buildFilterConvsMenu()
 
 	mainView = appkeys.New(mainViewNoKeys, mainKeyPress)
 
@@ -2878,11 +2971,40 @@ func Build() (*gowid.App, error) {
 		ChooseOne: &DarkMode,
 	}
 
-	appView = appkeys.New(
+	appViewWithKeys := appkeys.New(
 		assignTo(&appViewNoKeys, holder.New(mainView)),
 		appKeyPress,
 	)
 
+	Fin = rossshark.New(appViewWithKeys)
+
+	if !termshark.ConfBool("main.disable-shark-fin", false) {
+		steerableFin := appkeys.New(
+			Fin,
+			func(evk *tcell.EventKey, app gowid.IApp) bool {
+				if Fin.Active() {
+					switch evk.Key() {
+					case tcell.KeyLeft:
+						Fin.Dir = rossshark.Backward
+					case tcell.KeyRight:
+						Fin.Dir = rossshark.Forward
+					default:
+						Fin.Deactivate()
+					}
+					return true
+				}
+				return false
+			},
+			appkeys.Options{
+				ApplyBefore: true,
+			},
+		)
+
+		appView = holder.New(steerableFin)
+	} else {
+		appView = holder.New(appViewWithKeys)
+	}
+
 	// Create app, etc, but don't init screen which sets ICANON, etc
 	app, err = gowid.NewApp(gowid.AppArgs{
 		View:         appView,
@@ -2902,6 +3024,8 @@ func Build() (*gowid.App, error) {
 	app.RegisterMenu(analysisMenu)
 	app.RegisterMenu(generalMenu)
 	app.RegisterMenu(conversationMenu)
+	app.RegisterMenu(filterConvsMenu1)
+	app.RegisterMenu(filterConvsMenu2)
 
 	gowid.SetFocusPath(mainview, mainviewPaths[0], app)
 	gowid.SetFocusPath(altview1, altview1Paths[0], app)
