@@ -1,4 +1,4 @@
-// Copyright 2019 Graham Clark. All rights reserved.  Use of this source
+// Copyright 2019-2020 Graham Clark. All rights reserved.  Use of this source
 // code is governed by the MIT license that can be found in the LICENSE
 // file.
 
@@ -29,6 +29,7 @@ import (
 	"text/template"
 	"time"
 
+	"github.com/adam-hanna/arrayOperations"
 	"github.com/blang/semver"
 	"github.com/gcla/gowid"
 	"github.com/gcla/gowid/gwutil"
@@ -218,7 +219,11 @@ func TSharkVersion(tshark string) (semver.Version, error) {
 
 // Depends on empty.pcap being present
 func TSharkSupportsColor(tshark string) (bool, error) {
-	exitCode, err := RunForExitCode(tshark, "-r", CacheFile("empty.pcap"), "-T", "psml", "-w", os.DevNull, "--color")
+	exitCode, err := RunForExitCode(
+		tshark,
+		[]string{"-r", CacheFile("empty.pcap"), "-T", "psml", "--color"},
+		nil,
+	)
 	return exitCode == 0, err
 }
 
@@ -268,10 +273,15 @@ func TSharkPath() (string, *gowid.KeyValueError) {
 	return tsharkBin, nil
 }
 
-func RunForExitCode(prog string, args ...string) (int, error) {
+func RunForExitCode(prog string, args []string, env []string) (int, error) {
 	var err error
 	exitCode := -1 // default bad
 	cmd := exec.Command(prog, args...)
+	if env != nil {
+		cmd.Env = env
+	}
+	cmd.Stdout = ioutil.Discard
+	cmd.Stderr = ioutil.Discard
 	err = cmd.Run()
 	if err != nil {
 		if exerr, ok := err.(*exec.ExitError); ok {
@@ -317,6 +327,48 @@ func DumpcapBin() string {
 	return ConfString("main.dumpcap", "dumpcap")
 }
 
+func CapinfosBin() string {
+	return ConfString("main.capinfos", "capinfos")
+}
+
+// CaptureBin is the binary the user intends to use to capture
+// packets i.e. with the -i switch. This might be distinct from
+// DumpcapBin because dumpcap can't capture on extcap interfaces
+// like randpkt, but while tshark can, it can drop packets more
+// readily than dumpcap. This value is interpreted as the name
+// of a binary, resolved against PATH. Note that the default is
+// termshark - this invokes termshark in a special mode where it
+// first tries DumpcapBin, then if that fails, TSharkBin - for
+// the best of both worlds. To detect this, termshark will run
+// CaptureBin with TERMSHARK_CAPTURE_MODE=1 in the environment,
+// so when termshark itself is invoked with this in the environment,
+// it switches to capture mode.
+func CaptureBin() string {
+	if runtime.GOOS == "windows" {
+		return ConfString("main.capture-command", DumpcapBin())
+	} else {
+		return ConfString("main.capture-command", os.Args[0])
+	}
+}
+
+// PrivilegedBin returns a capture binary that may require setcap
+// privileges on Linux. This is a simple UI to cover the fact that
+// termshark's default capture method is to run dumpcap and tshark
+// as a fallback. I don't want to tell the user the capture binary
+// is termshark - that'd be confusing. We know that on Linux, termshark
+// will run dumpcap first, then fall back to tshark if needed. Only
+// dumpcap should need access to live interfaces; tshark is needed
+// for extcap interfaces only. This is used to provide advice to
+// the user if packet capture fails.
+func PrivilegedBin() string {
+	cap := CaptureBin()
+	if cap == "termshark" {
+		return DumpcapBin()
+	} else {
+		return cap
+	}
+}
+
 func TailCommand() []string {
 	def := []string{"tail", "-f", "-c", "+0"}
 	if runtime.GOOS == "windows" {
@@ -501,6 +553,25 @@ func SafePid(p IProcess) int {
 	return p.Pid()
 }
 
+func SetConvTypes(convs []string) {
+	SetConf("main.conv-types", convs)
+}
+
+func ConvTypes() []string {
+	defs := []string{"eth", "ip", "ipv6", "tcp", "udp"}
+	ctypes := ConfStrings("main.conv-types")
+	if len(ctypes) > 0 {
+		z, ok := arrayOperations.Intersect(defs, ctypes)
+		if ok {
+			res, ok := z.Interface().([]string)
+			if ok {
+				return res
+			}
+		}
+	}
+	return defs
+}
+
 func AddToRecentFiles(pcap string) {
 	comps := ConfStrings("main.recent-files")
 	if len(comps) == 0 || comps[0] != pcap {
@@ -664,7 +735,7 @@ var UnexpectedOutput = fmt.Errorf("Unexpected output")
 // Use tshark's output, becauses the indices can then be used to select
 // an interface to sniff on, and net.Interfaces returns the interfaces in
 // a different order.
-func Interfaces() (map[string]int, error) {
+func Interfaces() (map[int][]string, error) {
 	cmd := exec.Command(TSharkBin(), "-D")
 	out, err := cmd.Output()
 	if err != nil {
@@ -673,10 +744,10 @@ func Interfaces() (map[string]int, error) {
 	return interfacesFrom(bytes.NewReader(out))
 }
 
-func interfacesFrom(reader io.Reader) (map[string]int, error) {
-	re := regexp.MustCompile("^(?P<index>[0-9]+)\\.\\s+(?P<name1>[^\\s]+)(\\s*\\((?P<name2>[^)]+)\\))?")
+func interfacesFrom(reader io.Reader) (map[int][]string, error) {
+	re := regexp.MustCompile(`^(?P<index>[0-9]+)\.\s+(?P<name1>[^\s]+)(\s*\((?P<name2>[^)]+)\))?`)
 
-	res := make(map[string]int)
+	res := make(map[int][]string)
 	scanner := bufio.NewScanner(reader)
 	for scanner.Scan() {
 		line := scanner.Text()
@@ -687,7 +758,7 @@ func interfacesFrom(reader io.Reader) (map[string]int, error) {
 		}
 		result := make(map[string]string)
 		for i, name := range re.SubexpNames() {
-			if i != 0 && name != "" {
+			if i != 0 && match[i] != "" {
 				result[name] = match[i]
 			}
 		}
@@ -697,11 +768,13 @@ func interfacesFrom(reader io.Reader) (map[string]int, error) {
 			return nil, gowid.WithKVs(UnexpectedOutput, map[string]interface{}{"Output": line})
 		}
 
-		res[result["name1"]] = int(i)
+		val := make([]string, 0)
+		val = append(val, result["name1"])
 
 		if name2, ok := result["name2"]; ok {
-			res[name2] = int(i)
+			val = append([]string{name2}, val...)
 		}
+		res[int(i)] = val
 	}
 
 	return res, nil
@@ -933,6 +1006,20 @@ func CopyCommand(input io.Reader, cb interface{}) error {
 	return nil
 }
 
+// Returns true if error, too
+func FileSizeDifferentTo(filename string, cur int64) (int64, bool) {
+	var newSize int64
+	diff := true
+	fi, err := os.Stat(filename)
+	if err == nil {
+		newSize = fi.Size()
+		if cur == newSize {
+			diff = false
+		}
+	}
+	return newSize, diff
+}
+
 //======================================================================
 // Local Variables:
 // mode: Go
