@@ -1,4 +1,4 @@
-// Copyright 2019 Graham Clark. All rights reserved.  Use of this source
+// Copyright 2019-2020 Graham Clark. All rights reserved.  Use of this source
 // code is governed by the MIT license that can be found in the LICENSE
 // file.
 
@@ -98,9 +98,40 @@ func TestInterfaces1(t *testing.T) {
 `[1:]
 	interfaces, err := interfacesFrom(bytes.NewReader([]byte(out1)))
 	assert.NoError(t, err)
-	assert.Equal(t, 12, len(interfaces))
-	assert.Equal(t, interfaces[`\Device\NPF_{78032B7E-4968-42D3-9F37-287EA86C0AAA}`], 2)
-	assert.Equal(t, interfaces[`Local Area Connection* 10`], 2)
+	assert.Equal(t, 6, len(interfaces))
+	v := interfaces[2]
+	assert.Equal(t, `\Device\NPF_{78032B7E-4968-42D3-9F37-287EA86C0AAA}`, v[1])
+	assert.Equal(t, `Local Area Connection* 10`, v[0])
+}
+
+func TestInterfaces2(t *testing.T) {
+	out1 := `
+1. eth0
+2. ham0
+3. docker0
+4. vethd45103d
+5. lo (Loopback)
+6. mpqemubr0-dummy
+7. nflog
+8. nfqueue
+9. bluetooth0
+10. virbr0-nic
+11. vboxnet0
+12. ciscodump (Cisco remote capture)
+13. dpauxmon (DisplayPort AUX channel monitor capture)
+14. randpkt (Random packet generator)
+15. sdjournal (systemd Journal Export)
+16. sshdump (SSH remote capture)
+17. udpdump (UDP Listener remote capture)
+`[1:]
+	interfaces, err := interfacesFrom(bytes.NewReader([]byte(out1)))
+	assert.NoError(t, err)
+	assert.Equal(t, 17, len(interfaces))
+	v := interfaces[3]
+	assert.Equal(t, `docker0`, v[0])
+	v = interfaces[12]
+	assert.Equal(t, `Cisco remote capture`, v[0])
+	assert.Equal(t, `ciscodump`, v[1])
 }
 
 func TestConv1(t *testing.T) {
