@@ -1,4 +1,4 @@
-// Copyright 2019 Graham Clark. All rights reserved.  Use of this source
+// Copyright 2019-2020 Graham Clark. All rights reserved.  Use of this source
 // code is governed by the MIT license that can be found in the LICENSE
 // file.
 
@@ -43,7 +43,9 @@ import (
 	"github.com/gcla/termshark/v2/widgets/appkeys"
 	"github.com/gcla/termshark/v2/widgets/copymodetable"
 	"github.com/gcla/termshark/v2/widgets/framefocus"
+	"github.com/gcla/termshark/v2/widgets/keepselected"
 	"github.com/gcla/termshark/v2/widgets/regexstyle"
+	"github.com/gcla/termshark/v2/widgets/scrollabletable"
 	"github.com/gcla/termshark/v2/widgets/trackfocus"
 	"github.com/gcla/termshark/v2/widgets/withscrollbar"
 	"github.com/gdamore/tcell"
@@ -109,7 +111,7 @@ type Widget struct {
 	stats          streamStats             // track turns, client packets, bytes, etc
 	streamHeader   streams.FollowHeader    // first chunk of data back from tshark -z follow
 	displayAs      DisplayFormat           // display as hex, ascii, raw
-	captureDevice  string                  // gcla later todo - custom widget, don't make callbacks, who cares
+	captureDevice  string                  // it's a very feature-specific widget so I don't care about supporting callbacks
 	displayFilter  string                  // "tcp.stream eq 1"
 	Proto          streams.Protocol        // TCP, UDP
 	tableHolder    *holder.Widget          // hold the chunk UI table
@@ -239,7 +241,7 @@ func MakeConvMenu() (*holder.Widget, *menu.Widget) {
 
 func (w *Widget) updateConvMenuWidget(app gowid.IApp) {
 	convListBox := w.makeConvMenuWidget()
-	w.convMenuHolder.SetSubWidget(convListBox, app) // gcla later todo
+	w.convMenuHolder.SetSubWidget(convListBox, app)
 	w.setConvButtonText(app)
 	w.setTurnText(app)
 }
@@ -305,13 +307,14 @@ func makeTable(data chunkList) (gowid.IWidget, *copymodetable.Widget) {
 		"streamtable",
 		copyModePalette{},
 	)
-	sc := &keepSelected{
-		sub: withscrollbar.New(scrollableTableWidget{
-			Widget: cmtbl,
-		}, withscrollbar.Options{
-			HideIfContentFits: true,
-		}),
-	}
+	sc := keepselected.New(
+		withscrollbar.New(
+			scrollabletable.New(cmtbl),
+			withscrollbar.Options{
+				HideIfContentFits: true,
+			},
+		),
+	)
 
 	return sc, cmtbl
 }
@@ -535,7 +538,6 @@ func (w *Widget) construct() {
 	)
 
 	searchBoxStyled := styled.New(searchBox2,
-		// gcla later todo
 		gowid.MakePaletteRef("stream-search"),
 	)
 
@@ -551,7 +553,6 @@ func (w *Widget) construct() {
 	}
 
 	streamsFooter1 := columns.NewWithDim(
-		//gowid.RenderWithWeight{1},
 		fixed,
 		hpad,
 		hpadding.New(
@@ -568,8 +569,6 @@ func (w *Widget) construct() {
 		},
 		hpad,
 		hpadding.New(
-			//styled.NewInvertedFocus(nextBtn, gowid.MakePaletteRef("default")),
-			//styled.NewExt(nextBtn, gowid.MakePaletteRef("default"), gowid.
 			clicktracker.New(
 				styled.NewExt(
 					nextBtn,
@@ -1162,7 +1161,6 @@ func (c rawChunkList) CellWidgets(row table.RowId) []gowid.IWidget {
 	return res
 }
 
-// gcla later todo - duplicate of below, fix
 func (c asciiChunkList) Widths() []gowid.IWidgetDimension {
 	return []gowid.IWidgetDimension{gowid.RenderWithWeight{W: 1}}
 }
@@ -1204,21 +1202,6 @@ func (c chunkList) RowIdentifier(row int) (table.RowId, bool) {
 
 //======================================================================
 
-// To implement withscrollbar.IScrollValues
-type scrollableTableWidget struct {
-	*copymodetable.Widget
-}
-
-func (s scrollableTableWidget) ScrollLength() int {
-	return s.Model().(table.IBoundedModel).Rows()
-}
-
-func (s scrollableTableWidget) ScrollPosition() int {
-	return s.CurrentRow()
-}
-
-//======================================================================
-
 // TODO - duplicated from termshark
 
 type copyModePalette struct{}
@@ -1331,39 +1314,6 @@ func newData(clicker IChunkClicked, ca iClickIsActive, mapper iMapChunkToTableRo
 	return res
 }
 
-//======================================================================
-
-// A widget to ensure that its subwidget is always rendered as "selected", even if it's
-// not in focus. This allows a composite widget to style its selected child even without
-// focus so the user can see which child is active.
-type keepSelected struct {
-	sub gowid.IWidget
-}
-
-func (w *keepSelected) Render(size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) gowid.ICanvas {
-	return w.sub.Render(size, focus.SelectIf(true), app)
-}
-
-func (w *keepSelected) RenderSize(size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) gowid.IRenderBox {
-	return w.sub.RenderSize(size, focus.SelectIf(true), app)
-}
-
-func (w *keepSelected) UserInput(ev interface{}, size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) bool {
-	return w.sub.UserInput(ev, size, focus.SelectIf(true), app)
-}
-
-func (w *keepSelected) Selectable() bool {
-	return w.sub.Selectable()
-}
-
-func (w *keepSelected) SubWidget() gowid.IWidget {
-	return w.sub
-}
-
-func (w *keepSelected) SetSubWidget(wi gowid.IWidget, app gowid.IApp) {
-	w.sub = wi
-}
-
 //======================================================================
 // Local Variables:
 // mode: Go
