@@ -1,4 +1,4 @@
-// Copyright 2019 Graham Clark. All rights reserved.  Use of this source
+// Copyright 2019-2020 Graham Clark. All rights reserved.  Use of this source
 // code is governed by the MIT license that can be found in the LICENSE
 // file.
 
@@ -8,8 +8,11 @@ package withscrollbar
 import (
 	"github.com/gcla/gowid"
 	"github.com/gcla/gowid/widgets/columns"
+	"github.com/gcla/gowid/widgets/list"
 	"github.com/gcla/gowid/widgets/selectable"
+	"github.com/gcla/gowid/widgets/table"
 	"github.com/gcla/gowid/widgets/vscroll"
+	"github.com/gdamore/tcell"
 )
 
 //======================================================================
@@ -18,8 +21,10 @@ type Widget struct {
 	always   *columns.Widget // use if scrollbar is to be shown
 	w        IScrollSubWidget
 	sb       *vscroll.Widget
-	goUpDown int // positive means down
-	pgUpDown int // positive means down
+	goUpDown int     // positive means down
+	pgUpDown int     // positive means down
+	frac     float32 // positive means down
+	fracSet  bool
 	opt      Options
 }
 
@@ -45,6 +50,14 @@ type IScrollOnePage interface {
 	DownPage(num int, size gowid.IRenderSize, app gowid.IApp)
 }
 
+type IScrollHome interface {
+	GoHome(size gowid.IRenderSize, app gowid.IApp)
+}
+
+type IScrollToEnd interface {
+	GoToEnd(size gowid.IRenderSize, app gowid.IApp)
+}
+
 type IScrollSubWidget interface {
 	gowid.IWidget
 	IScrollValues
@@ -72,14 +85,13 @@ func New(w IScrollSubWidget, opts ...Options) *Widget {
 				D:       gowid.RenderWithUnits{U: 1},
 			},
 		}),
-		w:        w,
-		sb:       sb,
-		goUpDown: 0,
-		pgUpDown: 0,
-		opt:      opt,
+		w:   w,
+		sb:  sb,
+		opt: opt,
 	}
 	sb.OnClickAbove(gowid.MakeWidgetCallback("cb", res.clickUp))
 	sb.OnClickBelow(gowid.MakeWidgetCallback("cb", res.clickDown))
+	sb.OnRightClick(gowid.MakeWidgetCallbackExt("cb", res.rightClick))
 	sb.OnClickUpArrow(gowid.MakeWidgetCallback("cb", res.clickUpArrow))
 	sb.OnClickDownArrow(gowid.MakeWidgetCallback("cb", res.clickDownArrow))
 	return res
@@ -93,6 +105,12 @@ func (e *Widget) clickDown(app gowid.IApp, w gowid.IWidget) {
 	e.pgUpDown += 1
 }
 
+func (e *Widget) rightClick(app gowid.IApp, w gowid.IWidget, data ...interface{}) {
+	frac := data[0].(float32)
+	e.frac = frac
+	e.fracSet = true
+}
+
 func (e *Widget) clickUpArrow(app gowid.IApp, w gowid.IWidget) {
 	e.goUpDown -= 1
 }
@@ -130,6 +148,39 @@ func (w *Widget) UserInput(ev interface{}, size gowid.IRenderSize, focus gowid.S
 	w.sb.Middle = y
 	w.sb.Bottom = z
 
+	if ws, ok := w.w.(IScrollOnePage); ok {
+		if ev, ok := ev.(*tcell.EventKey); ok {
+			switch ev.Key() {
+			case tcell.KeyPgUp:
+				ws.UpPage(1, size, app)
+				return true
+			case tcell.KeyPgDn:
+				ws.DownPage(1, size, app)
+				return true
+			}
+		}
+	}
+
+	if ws, ok := w.w.(IScrollHome); ok {
+		if ev, ok := ev.(*tcell.EventKey); ok {
+			switch ev.Key() {
+			case tcell.KeyHome:
+				ws.GoHome(size, app)
+				return true
+			}
+		}
+	}
+
+	if ws, ok := w.w.(IScrollToEnd); ok {
+		if ev, ok := ev.(*tcell.EventKey); ok {
+			switch ev.Key() {
+			case tcell.KeyEnd:
+				ws.GoToEnd(size, app)
+				return true
+			}
+		}
+	}
+
 	res := w.always.UserInput(ev, size, focus, app)
 	if res {
 		w.always.SetFocus(app, 0)
@@ -137,6 +188,10 @@ func (w *Widget) UserInput(ev interface{}, size gowid.IRenderSize, focus gowid.S
 	return res
 }
 
+type iSetPosition interface {
+	SetPos(pos list.IBoundedWalkerPosition, app gowid.IApp)
+}
+
 func (w *Widget) Render(size gowid.IRenderSize, focus gowid.Selector, app gowid.IApp) gowid.ICanvas {
 	if w.opt.HideIfContentFits && w.contentFits(size) {
 		return w.w.Render(size, focus, app)
@@ -152,7 +207,7 @@ func (w *Widget) Render(size gowid.IRenderSize, focus gowid.Selector, app gowid.
 	var z int
 	if ecols >= 1 {
 		ebox := gowid.MakeRenderBox(ecols, box.BoxRows())
-		if w.goUpDown != 0 || w.pgUpDown != 0 {
+		if w.goUpDown != 0 || w.pgUpDown != 0 || w.fracSet {
 			if w.goUpDown > 0 {
 				w.w.Down(w.goUpDown, ebox, app)
 			} else if w.goUpDown < 0 {
@@ -164,6 +219,13 @@ func (w *Widget) Render(size gowid.IRenderSize, focus gowid.Selector, app gowid.
 			} else if w.pgUpDown < 0 {
 				w.w.UpPage(-w.pgUpDown, ebox, app)
 			}
+
+			if w.fracSet {
+				if wp, ok := w.w.(iSetPosition); ok {
+					wp.SetPos(table.Position(int(float32(w.w.ScrollLength()-1)*w.frac)), app)
+				}
+				w.fracSet = false
+			}
 		}
 		w.goUpDown = 0
 		w.pgUpDown = 0
